#if defined(__linux__)

#if defined(CROSS)
/* To avoid compile error by differ of return value of read()/write() */
#define _READ_WRITE_RETURN_TYPE _ssize_t
#endif
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/ioctl.h>

#include <nllibc.h>

int _set_errno(int retval)
{
  if ((retval < 0) && (retval > -4096)) {
    errno = -retval;
    retval = -1;
  }
  return retval;
}

struct rusage;

void ___exit(int status);
pid_t __fork(void);
pid_t __wait4(pid_t wpid, int *status, int options, struct rusage *rusage);
int __kill(pid_t pid, int sig);
int __execve(const char *path, char *const argv[], char *const envp[]);

ssize_t __read(int fd, void *buf, size_t nbytes);
ssize_t __write(int fd, const void *buf, size_t nbytes);
int __open(const char *path, int flags, mode_t mode);
int __close(int fd);

int __dup(int oldd);
int __pipe(int fildes[2]);
int __dup2(int oldd, int newd);
int __pipe2(int fildes[2], int flags);

int __sigaction(int sig, const struct sigaction *act, struct sigaction *oact);
int __sigprocmask(int how, const sigset_t *set, sigset_t *oset);

int __chdir(const char *path);
int __ioctl(int fd, unsigned long request, unsigned long args);
int __brk(const void *addr);
int ____getcwd(char *buf, size_t size);

void _exit(int status)
{
  ___exit(status);
  while (1) /* to avoid noreturn warning */
    ;
}

pid_t fork(void)
{
  return _set_errno(__fork());
}

pid_t wait4(pid_t wpid, int *status, int options, struct rusage *rusage)
{
  return _set_errno(__wait4(wpid, status, options, rusage));
}

int kill(pid_t pid, int sig)
{
  return _set_errno(__kill(pid, sig));
}

int execve(const char *path, char *const argv[], char *const envp[])
{
  return _set_errno(__execve(path, argv, envp));
}

ssize_t read(int fd, void *buf, size_t nbytes)
{
  return _set_errno(__read(fd, buf, nbytes));
}

ssize_t write(int fd, const void *buf, size_t nbytes)
{
  return _set_errno(__write(fd, buf, nbytes));
}

static int _va_open(const char *path, int flags, va_list ap)
{
  unsigned long arg;
  arg = va_arg(ap, unsigned long);
  return __open(path, flags, arg);
}

int open(const char *path, int flags, ...)
{
  va_list arg;
  int r;

  va_start(arg, flags);
  r = _va_open(path, flags, arg);
  va_end(arg);

  return _set_errno(r);
}

int close(int fd)
{
  return _set_errno(__close(fd));
}

int dup(int oldd)
{
  return _set_errno(__dup(oldd));
}

int pipe(int fildes[2])
{
  return _set_errno(__pipe(fildes));
}

int dup2(int oldd, int newd)
{
  return _set_errno(__dup2(oldd, newd));
}

int pipe2(int fildes[2], int flags)
{
  return _set_errno(__pipe2(fildes, flags));
}

int sigaction(int sig, const struct sigaction *act, struct sigaction *oact)
{
  return _set_errno(__sigaction(sig, act, oact));
}

int sigprocmask(int how, const sigset_t *set, sigset_t *oset)
{
  return _set_errno(__sigprocmask(how, set, oset));
}

int chdir(const char *path)
{
  return _set_errno(__chdir(path));
}

static int _va_ioctl(int fd, unsigned long request, va_list ap)
{
  unsigned long arg;
  arg = va_arg(ap, unsigned long);
  return __ioctl(fd, request, arg);
}

int ioctl(int fd, unsigned long request, ...)
{
  va_list arg;
  int r;

  va_start(arg, request);
  r = _va_ioctl(fd, request, arg);
  va_end(arg);

  return _set_errno(r);
}

int brk(const void *addr)
{
  return _set_errno(__brk(addr));
}

int __getcwd(char *buf, size_t size)
{
  return _set_errno(____getcwd(buf, size));
}
#endif
